import os

import pyarrow as pa
import pyarrow.compute as pc
import pyarrow.fs as pfs
import pyarrow.parquet as pq
from fsspec import AbstractFileSystem

from .helpers.misc import read_table, run_parallel


def sort_schema(schema: pa.Schema, names: list[str] | None = None) -> pa.Schema:
    """Sort fields of a pyarrow schema in alphabetical order.

    Args:
        schema (pa.Schema): Pyarrow schema.

    Returns:
        pa.Schema: Sorted pyarrow schema.
    """
    if names is not None:
        names = names + [name for name in schema.names if name not in names]
        names = [name for name in names if name in schema.names]
    else:
        names = schema.names

    return pa.schema(
        [pa.field(name, type_) for name, type_ in sorted(zip(names, schema.types))]
    )


def shrink_large_string(schema: pa.Schema) -> pa.Schema:
    """Convert all large_string types to string in a pyarrow.schema.

    Args:
        schema (pa.Schema): pyarrow schema

    Returns:
        pa.Schema: converted pyarrow.schema
    """
    return pa.schema(
        [
            (n, pa.utf8()) if t == pa.large_string() else (n, t)
            for n, t in list(zip(schema.names, schema.types))
        ]
    )


def convert_timestamp(
    schema: pa.Schema,
    timestamp_fields: str | list[str] | None = None,
    unit: str | None = "us",
    tz: str | None = None,
    remove_tz: bool = False,
) -> pa.Schema:
    """Convert timestamp based on given unit and timezone.

    Args:
        schema (pa.Schema): Pyarrow schema.
        timestamp_fields (str | list[str] | None, optional): timestamp fields. Defaults to None.
        unit (str | None): timestamp unit. Defaults to "us".
        tz (None, optional): timezone. Defaults to None.
        remove_tz (str | None): Wheter to remove the timezone from the timestamp or not. Defaults to False.

    Returns:
        pa.Schema: Pyarrow schema with converted timestamp(s).
    """
    if timestamp_fields is None:
        timestamp_fields = [
            f
            for f in schema.names
            if isinstance(schema.field(f).type, pa.TimestampType)
        ]

    if isinstance(timestamp_fields, str):
        timestamp_fields = [timestamp_fields]

    for timestamp_field in timestamp_fields:
        timestamp_field_idx = schema.get_field_index(timestamp_field)
        unit = unit or schema.field(timestamp_field).type.unit
        tz = tz or schema.field(timestamp_field).type.tz
        if remove_tz:
            tz = None
        schema = schema.remove(timestamp_field_idx).insert(
            timestamp_field_idx,
            pa.field(timestamp_field, pa.timestamp(unit=unit, tz=tz)),
        )

    return schema


def replace_field(schema: pa.Schema, field: str, dtype: pa.DataType) -> pa.Schema:
    """Replace the dtype of a field in a pyarrow schema.

    Args:
        schema (pa.Schema): Pyarrow schema.
        field (str): Field name.
        dtype (pa.DataType): Pyarrow data type.

    Returns:
        pa.Schema: Pyarrow schema with replaced dtype.
    """
    if field in schema.names:
        field_idx = schema.get_field_index(field)
        return schema.remove(field_idx).insert(field_idx, pa.field(field, dtype))


def append_field(schema: pa.Schema, field: str, dtype: pa.DataType) -> pa.Schema:
    """
    Append a new field to the given schema.

    Args:
        schema (pa.Schema): The schema to append the field to.
        field (str): The name of the field to append.
        dtype (pa.DataType): The data type of the field to append.

    Returns:
        pa.Schema: The updated schema with the new field appended.
    """
    if field not in schema.names:
        return schema.append(pa.field(field, dtype))
    return schema


def modify_field(schema: pa.Schema, field: str, dtype: pa.DataType) -> pa.Schema:
    """
    Modify a field in the given schema.

    Args:
        schema (pa.Schema): The schema to modify the field in.
        field (str): The name of the field to modify.
        dtype (pa.DataType): The data type of the field to modify.

    Returns:
        pa.Schema: The updated schema with the field modified.
    """
    if field in schema.names:
        return replace_field(schema, field, dtype)
    return append_field(schema, field, dtype)


def cast_int2timestamp(
    table: pa.Table, column: str | list[str], unit: str = "us", tz: str | None = None
) -> pa.Table:
    column = [column] if isinstance(column, str) else column
    columns = table.schema.names
    for col in column:
        arr = table.column(col)
        arr_new = pa.array(arr.to_pylist(), pa.timestamp(unit=unit, tz=tz))
        table = table.drop(col).append_column(col, arr_new)

    return table.select(columns)


def cast_str2bool(
    table: pa.Table,
    schema: pa.Schema | None = None,
    column: str | list[str] | None = None,
    true_values: list[str] = [
        "true",
        "wahr",
        "1",
        "1.0",
        "yes",
        "ja",
        "ok",
        "o.k",
        "okay",
    ],
    skip_nulls: bool = False,
) -> pa.Table:
    if column is None and schema is not None:
        column = [
            col
            for col in table.schema.names
            if pa.types.is_boolean(schema.field(col).type)
            and pa.types.is_string(table.schema.field(col).type)
        ]
    column = [column] if isinstance(column, str) else column
    columns = table.schema.names
    for col in column:
        arr = table.column(col)
        arr_is_in = pc.is_in(
            pc.utf8_lower(table.column(col)),
            pa.array(true_values),
            skip_nulls=skip_nulls,
        )
        arr_is_null = pc.is_null(arr)
        arr_new = pc.if_else(arr_is_null, None, arr_is_in)
        table = table.drop(col).append_column(col, arr_new)
    return table.select(columns)


def replace_schema(
    table: pa.Table,
    schema: pa.Schema | None = None,
    field_dtype: dict | None = None,
    ts_unit: str | None = "us",
    tz: str | None = None,
    alter_schema: bool = True,
) -> pa.Table:
    """
    Replaces the schema of a PyArrow table with the specified schema or modifies the existing schema.

    Args:
        table (pa.Table): The PyArrow table to modify.
        schema (pa.Schema | None, optional): The new schema to replace the existing schema. Defaults to None.
        field_dtype (dict | None, optional): A dictionary mapping field names to their new data types. Defaults to None.
        alter_schema (bool, optional): If True, adds missing fields to the table based on the new schema.
            If False, drops fields from the table that are not present in the new schema. Defaults to True.

    Returns:
        pa.Table: The modified table with the replaced or modified schema.
    """
    schema_org = table.schema
    if field_dtype is not None:
        schema = schema or table.schema
        for field, dtype in field_dtype.items():
            schema = modify_field(schema=schema, field=field, dtype=dtype)

    schema = schema or table.schema

    if schema == schema_org:
        return table

    # add missing fields to the table
    missing_fields = [
        field for field in schema.names if field not in table.column_names
    ]

    for field in missing_fields:
        table = table.append_column(
            field, pa.array([None] * len(table), type=schema.field(field).type)
        )

    if not alter_schema:
        table = table.drop(
            [field for field in table.column_names if field not in schema.names]
        )

    int2timestamp_columns = [
        col
        for col in schema.names
        if pa.types.is_timestamp(schema.field(col).type)
        and (
            pa.types.is_integer(schema_org.field(col).type)
            if col in schema_org.names
            else False
        )
    ]
    table = cast_int2timestamp(table, int2timestamp_columns, unit=ts_unit, tz=tz)
    # cast str to bool
    table = cast_str2bool(table, schema=schema)

    return table.select(schema.names).cast(schema)

    return table.select(schema.names)


# def _unify_schemas(
#     schema1: pa.Schema,
#     schema2: pa.Schema,
#     ts_unit: str | None = None,  # "us",
#     tz: str | None = None,
#     use_large_string: bool = False,
#     sort: bool | list[str] | str = False,
# ) -> tuple[dict, bool]:
#     """Returns a unified pyarrow schema.

#     Args:
#         schema1 (pa.Schema): Pyarrow schema 1.
#         schema2 (pa.Schema): Pyarrow schema 2.
#         ts_unit (str|None): timestamp unit.
#         tz (str|None): timezone for timestamp fields. Defaults to "UTC".
#         use_large_string (bool): Convert pyarrow.large_string() to pyarrow.string().
#             Defaults to False.
#         sort (bool | list[str]): Use this, to define the field(column) order. If it is
#             True, the fields are sorted alphanumerically, if it is False, the field
#             order of firt schema is used. Otherwise, the order if the given field names is
#             applied.
#     Returns:
#         Tuple[dict, bool]: Unified pyarrow schema, bool value if schemas were equal.
#     """

#     dtypes = [
#         pa.null(),
#         pa.int8(),
#         pa.int16(),
#         pa.int32(),
#         pa.int64(),
#         pa.float16(),
#         pa.float32(),
#         pa.float64(),
#         pa.string(),
#         pa.utf8(),
#         pa.large_string(),
#         pa.large_utf8(),
#     ]
#     timestamp_units = ["ns", "us", "ms", "s"]
#     # string_dtypes = [pa.string(), pa.utf8(), pa.large_string(), pa.large_utf8()]

#     # check for equal columns and column order
#     if schema1.names == schema2.names:
#         # if schema1.types == schema2.types:
#         #    return schema1, True

#         all_names = schema1.names
#         file_schemas_equal = True

#     elif sorted(schema1.names) == sorted(schema2.names):
#         all_names = schema1.names
#         file_schemas_equal = False

#     else:
#         # all_names = list(set(schema1.names + schema2.names))
#         all_names = schema1.names + [
#             name for name in schema2.names if name not in schema1.names
#         ]
#         file_schemas_equal = False

#     if sort:
#         if isinstance(sort, bool):
#             all_names = sorted(all_names)
#         elif isinstance(sort, list):
#             sort = [name for name in sort if name in all_names]
#             all_names = sort + [name for name in all_names if name not in sort]

#     schema = []
#     for name in all_names:
#         if name in schema1.names:
#             type1 = schema1.field(name).type
#         else:
#             type1 = schema2.field(name).type
#         if name in schema2.names:
#             type2 = schema2.field(name).type
#         else:
#             type2 = schema1.field(name).type

#         if type1 != type2:
#             if type1 in dtypes:
#                 rank1 = dtypes.index(type1) if type1 in dtypes else 0
#                 rank2 = dtypes.index(type2) if type2 in dtypes else 0

#                 if type1.num_buffers == 2 and type2.num_buffers == 2:
#                     bit_width1 = type1.bit_width
#                     bit_width2 = type2.bit_width
#                     if bit_width1 > bit_width2:
#                         if rank1 > rank2:
#                             type_ = type1
#                         else:
#                             type_ = pa.float64()
#                         # (
#                         #     pa.float16()
#                         #     if bit_width1 == 16
#                         #     else pa.float32()
#                         #     if bit_width1 == 32
#                         #     else pa.float64()
#                         # )
#                     else:
#                         if rank2 > rank1:
#                             type_ = type2
#                         else:
#                             type_ = pa.float64()
#                 else:
#                     type_ = type1 if rank1 > rank2 else type2

#             elif isinstance(type1, pa.TimestampType):
#                 rank1 = (
#                     timestamp_units.index(type1.unit)
#                     if isinstance(type1, pa.TimestampType)
#                     else 0
#                 )
#                 if isinstance(type2, pa.TimestampType):
#                     rank2 = timestamp_units.index(type2.unit)
#                 else:
#                     rank2 = 0
#                 type_ = type1 if rank1 > rank2 else type2
#             else:
#                 rank1 = 1
#                 rank2 = 0
#                 type_ = type1 if rank1 > rank2 else type2
#             schema.append(pa.field(name, type_))

#         else:
#             schema.append(pa.field(name, type1))

#     schema = pa.schema(schema)

#     if ts_unit is not None or tz is not None:
#         schema = convert_timestamp(schema, unit=ts_unit, tz=tz)

#     if not use_large_string:
#         schema = shrink_large_string(schema)

#     if schema != schema1 or schema != schema2:
#         file_schemas_equal = False

#     return schema, file_schemas_equal


# def unify_schemas(
#     schemas: list[pa.Schema],
#     ts_unit: str | None = "us",
#     tz: str | None = None,
#     use_large_string: bool = False,
#     sort: bool | list[str] = False,
# ) -> tuple[pa.Schema, bool]:
#     """Get the unified pyarrow schema for a list of schemas.

#     Args:
#         schemas (list[pa.Schema]): Pyarrow schemas.
#         ts_unit (str|None): timestamp unit.
#         tz (str|None): timezone for timestamp fields. Defaults to "UTC".
#         use_large_string (bool): Convert pyarrow.large_string() to pyarrow.string().
#             Defaults to False.
#         sort (bool | list[str]): Use this, to define the field(column) order. If it is
#             True, the fields are sorted alphanumerically, if it is False, the field
#             order of firt schema is used. Otherwise, the order if the given field names is
#             applied.
#     Returns:
#         tuple[pa.Schema, bool]: Unified pyarrow schema.
#     """

#     schemas_equal = True
#     unified_schema = schemas[0]
#     for schema in schemas[1:]:
#         unified_schema, schemas_equal_ = _unify_schemas(
#             unified_schema,
#             schema,
#             ts_unit=ts_unit,
#             tz=tz,
#             use_large_string=use_large_string,
#             sort=sort,
#         )

#         schemas_equal *= schemas_equal_

#     return unified_schema, bool(schemas_equal)


def repair_schema(
    files: list[str] | None = None,
    file_schemas: dict[str, pa.Schema] | None = None,
    base_path: str | None = None,
    schema: pa.Schema | None = None,
    filesystem: AbstractFileSystem | pfs.FileSystem | None = None,
    n_jobs: int = -1,
    backend: str = "threading",
    verbose: bool = True,
    ts_unit: str | None = None,  # "us",
    tz: str | None = None,
    use_large_string: bool = False,
    # sort: bool | list[str] = False,
    alter_schema: bool = True,
    **kwargs,
):
    """Repairs the pyarrow schema of a parquet or arrow dataset

    Args:
        files (list[str]): Files of the dataset.
        schema (pa.Schema | None, optional): Unified pyarrow schema.
            If None, the unified schema is generated from the given files of the dataset. Defaults to None.
        filesystem (AbstractFileSystem | pfs.FileSystem | None, optional): Filesystem. Defaults to None.
        n_jobs (int, optional): n_jobs parameter of joblib.Parallel. Defaults to -1.
        backend (str, optional): backend parameter of joblib.Parallel. Defaults to "threading".
        verbose (bool, optional): Wheter to show the task progress using tqdm or not. Defaults to True.
        ts_unit (str|None): timestamp unit.
        tz (str|None): timezone for timestamp fields. Defaults to "UTC".
        use_large_string (bool): Convert pyarrow.large_string() to pyarrow.string().
            Defaults to False.
        sort (bool | list[str]): Use this, to define the field(column) order. If it is
            True, the fields are sorted alphanumerically, if it is False, the field
            order of firt schema is used. Otherwise, the order if the given field names is
            applied.
    """
    if files is None:
        if file_schemas is not None:
            files = list(file_schemas.keys())
        else:
            raise ValueError("Either files or file_schemas must be provided.")

    if file_schemas is None:
        file_schemas = collect_file_schemas(
            files=files,
            base_path=base_path,
            filesystem=filesystem,
            n_jobs=n_jobs,
            backend=backend,
            verbose=verbose,
        )

    if base_path is not None:
        files = [os.path.join(base_path, f) for f in files]

    if schema is None:
        schema, schemas_equal = pa.unify_schemas(
            schemas=list(file_schemas.values()),
        )

    if ts_unit is not None or tz is not None:
        schema = convert_timestamp(schema, unit=ts_unit, tz=tz)

    if not use_large_string:
        schema = shrink_large_string(schema)

    schemas_equal = all([schema == schemas_ for schemas_ in file_schemas])

    if not schemas_equal:
        files = [f for f in files if file_schemas[f] != schema]

    def _repair_schema(f, schema, filesystem):
        if file_schemas is not None:
            file_schema = file_schemas.get(
                f.replace(base_path or "", ""), None
            ) or pq.read_schema(f, filesystem=filesystem)
        else:
            file_schema = None

        if file_schema is None:
            file_schema = pq.read_schema(f, filesystem=filesystem)

        if file_schema != schema:
            table = replace_schema(
                read_table(f, filesystem=filesystem, partitioning=None),
                schema=schema,
                ts_unit=ts_unit,
                tz=tz,
                alter_schema=alter_schema,
            )
            pq.write_table(
                table,
                f,
                filesystem=filesystem,
                coerce_timestamps=ts_unit,
                allow_truncated_timestamps=True,
                **kwargs,
            )

    _ = run_parallel(
        _repair_schema,
        files,
        schema=schema,
        filesystem=filesystem,
        backend=backend,
        n_jobs=n_jobs,
        verbose=verbose,
    )


def collect_file_schemas(
    files: list[str] | str,
    base_path: str | None = None,
    filesystem: AbstractFileSystem | pfs.FileSystem | None = None,
    n_jobs: int = -1,
    backend: str = "threading",
    verbose: bool = True,
) -> dict[str, pa.Schema]:
    """Collect all pyarrow schemas for the given parquet files.

    Args:
        files (list[str] | str): Parquet files.
        filesystem (AbstractFileSystem | pfs.FileSystem | None, optional): Filesystem. Defaults to None.
        n_jobs (int, optional): n_jobs parameter of joblib.Parallel. Defaults to -1.
        backend (str, optional): backend parameter of joblib.Parallel. Defaults to "threading".
        verbose (bool, optional): Wheter to show the task progress using tqdm or not. Defaults to True.

    Returns:
        dict[str, pa.Schema]: Pyarrow schemas of the given files.
    """
    if base_path is not None:
        files = [os.path.join(base_path, f) for f in files]

    def get_schema(f, filesystem):
        return {f: pq.read_schema(f, filesystem=filesystem)}

    schemas = run_parallel(
        get_schema,
        files,
        filesystem=filesystem,
        backend=backend,
        n_jobs=n_jobs,
        verbose=verbose,
    )
    return {key: value for d in schemas for key, value in d.items()}



#############################################################################################################
# This was originally implemented by Burak Emre Kabakcı (@buremba on github) in its universql project
# project: https://github.com/buremba/universql
# original code: https://github.com/buremba/universql/blob/main/universql/lake/fsspec_util.py
#############################################################################################################


import inspect
import os
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path

import psutil
from fsspec.implementations.cache_mapper import AbstractCacheMapper
from fsspec.implementations.cached import SimpleCacheFileSystem
from loguru import logger


def get_total_directory_size(directory: str):
    return sum(f.stat().st_size for f in Path(directory).glob("**/*") if f.is_file())


class FileNameCacheMapper(AbstractCacheMapper):
    def __init__(self, directory):
        self.directory = directory

    def __call__(self, path: str) -> str:
        os.makedirs(os.path.dirname(os.path.join(self.directory, path)), exist_ok=True)
        return path


class throttle(object):
    """
    Decorator that prevents a function from being called more than once every
    time period.
    To create a function that cannot be called more than once a minute:
        @throttle(minutes=1)
        def my_fun():
            pass
    """

    def __init__(self, seconds=0, minutes=0, hours=0):
        self.throttle_period = timedelta(seconds=seconds, minutes=minutes, hours=hours)
        self.time_of_last_call = datetime.min

    def __call__(self, fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            now = datetime.now()
            time_since_last_call = now - self.time_of_last_call

            if time_since_last_call > self.throttle_period:
                self.time_of_last_call = now
                return fn(*args, **kwargs)

        return wrapper


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


last_free = None
first_free = None


def get_friendly_disk_usage(storage: str) -> str:
    global last_free
    global first_free
    usage = psutil.disk_usage(storage)
    if first_free is None:
        first_free = usage.free
    current_usage = get_total_directory_size(storage)
    message = f"{sizeof_fmt(current_usage)} used {sizeof_fmt(usage.free)} available"
    if last_free is not None:
        downloaded_recently = last_free - usage.free
        if downloaded_recently > 10_000_000:
            downloaded_since_start = first_free - usage.free
            if downloaded_recently != downloaded_since_start:
                message += f" delta: {sizeof_fmt(downloaded_recently)}"
            message += f" delta since start: {sizeof_fmt(downloaded_since_start)}"

    last_free = usage.free
    return message


class MonitoredSimpleCacheFileSystem(SimpleCacheFileSystem):
    def __init__(self, **kwargs):
        kwargs["cache_storage"] = os.path.join(
            kwargs.get("cache_storage"), kwargs.get("fs").protocol[0]
        )
        super().__init__(**kwargs)
        self._mapper = FileNameCacheMapper(kwargs.get("cache_storage"))

    def _check_file(self, path):
        self._check_cache()
        cache_path = self._mapper(path)
        for storage in self.storage:
            fn = os.path.join(storage, cache_path)
            if os.path.exists(fn):
                return fn
            logger.info(f"Downloading {self.protocol[0]}://{path}")

    # def glob(self, path):
    #    return [self._strip_protocol(path)]

    def size(self, path):
        cached_file = self._check_file(self._strip_protocol(path))
        if cached_file is None:
            return self.fs.size(path)
        else:
            return os.path.getsize(cached_file)

    def __getattribute__(self, item):
        if item in {
            # new items
            "size",
            "glob",
            # previous
            "load_cache",
            "_open",
            "save_cache",
            "close_and_update",
            "__init__",
            "__getattribute__",
            "__reduce__",
            "_make_local_details",
            "open",
            "cat",
            "cat_file",
            "cat_ranges",
            "get",
            "read_block",
            "tail",
            "head",
            "info",
            "ls",
            "exists",
            "isfile",
            "isdir",
            "_check_file",
            "_check_cache",
            "_mkcache",
            "clear_cache",
            "clear_expired_cache",
            "pop_from_cache",
            "local_file",
            "_paths_from_path",
            "get_mapper",
            "open_many",
            "commit_many",
            "hash_name",
            "__hash__",
            "__eq__",
            "to_json",
            "to_dict",
            "cache_size",
            "pipe_file",
            "pipe",
            "start_transaction",
            "end_transaction",
        }:
            # all the methods defined in this class. Note `open` here, since
            # it calls `_open`, but is actually in superclass
            return lambda *args, **kw: getattr(type(self), item).__get__(self)(
                *args, **kw
            )
        if item in ["__reduce_ex__"]:
            raise AttributeError
        if item in ["transaction"]:
            # property
            return type(self).transaction.__get__(self)
        if item in ["_cache", "transaction_type"]:
            # class attributes
            return getattr(type(self), item)
        if item == "__class__":
            return type(self)
        d = object.__getattribute__(self, "__dict__")
        fs = d.get("fs", None)  # fs is not immediately defined
        if item in d:
            return d[item]
        elif fs is not None:
            if item in fs.__dict__:
                # attribute of instance
                return fs.__dict__[item]
            # attributed belonging to the target filesystem
            cls = type(fs)
            m = getattr(cls, item)
            if (inspect.isfunction(m) or inspect.isdatadescriptor(m)) and (
                not hasattr(m, "__self__") or m.__self__ is None
            ):
                # instance method
                return m.__get__(fs, cls)
            return m  # class method or attribute
        else:
            # attributes of the superclass, while target is being set up
            return super().__getattribute__(item)



import os

import duckdb
import pandas as pd
import pyarrow as pa
import yaml
from fsspec import AbstractFileSystem
from munch import Munch, munchify, toYAML, unmunchify

from pydala.helpers.polars import pl

from .dataset import CsvDataset, JsonDataset, ParquetDataset, PyarrowDataset
from .filesystem import FileSystem
from .helpers.misc import (delattr_rec, get_nested_keys, getattr_rec,
                           setattr_rec)
from .helpers.sql import get_table_names
from .table import PydalaTable


# @dataclass
class Catalog:
    def __init__(
        self,
        path: str,
        namespace: str | None = None,
        ddb_con: duckdb.DuckDBPyConnection | None = None,
        filesystem: AbstractFileSystem | None = None,
        bucket: str | None = None,
        **fs_kwargs,
    ):
        self._catalog_filesystem = FileSystem(bucket=bucket, fs=filesystem, **fs_kwargs)
        self._catalog_path = self._catalog_filesystem.expand_path(path)[0]
        self._namespace = namespace
        self.load_catalog(namespace=namespace)

        self.ddb_con = ddb_con
        if self.ddb_con is None:
            self.ddb_con = duckdb.connect()

        self._load_filesystems()
        self.table = {}

    def load_catalog(self, namespace: str | None = None) -> Munch:
        with self._catalog_filesystem.open(self._catalog_path, "r") as f:
            catalog = munchify(yaml.full_load(f))

        for ns in catalog.tables.keys():
            if catalog.tables[ns] is None:
                catalog.tables[ns] = Munch()

        self.params = catalog

        if namespace is not None:
            self.params.tables = self.params.tables[namespace]
        else:
            self.params.tables = self.params.tables

    def _write_catalog(self, delete_table: str = None):
        with self._catalog_filesystem.open(self._catalog_path, "r") as f:
            catalog = munchify(yaml.full_load(f))
        if delete_table is not None:
            delattr_rec(catalog, delete_table)

        if self._namespace is not None:
            self.params.tables[self._namespace].update(self.params.tables)
        else:
            self.params.tables.update(self.params.tables)

        catalog.update(self.params)

        with self._catalog_filesystem.open(self._catalog_path, "w") as f:
            yaml.dump(unmunchify(catalog), f)

    def _load_filesystems(self):
        if hasattr(self.params, "filesystem"):
            self.fs = Munch()

            for name in self.params.filesystem:
                if self.params.filesystem[name].protocol in ["file", "local"]:
                    self.params.filesystem[name].bucket = os.path.join(
                        os.path.dirname(self._catalog_path),
                        self.params.filesystem[name].bucket,
                    )
                fs = FileSystem(**self.params.filesystem[name])
                type(fs).protocol = name

                if self.params.filesystem[name].protocol != "file":
                    self.ddb_con.register_filesystem(fs)

                self.fs[name] = fs
        else:
            self.fs = None

    def _get_table_from_table_name(self, table_name: str) -> list[str]:
        table_name_items = table_name.split(".")

    # @staticmethod
    # def _get_table_from_table_name(self, table_name: str) -> str:
    #     table_name_items = table_name.split(".")

    #     if table_name_items[0] in self.list_namespaces:
    #         table_name_items.insert(1, "tables")
    #         return ".".join(table_name_items)

    #     if self._namespace is not None:
    #         table_name_items = [self._namespace, "tables"] + table_name_items
    #         return ".".join(table_name_items)

    #     return ".".join(table_name_items)

    def _get_table_params(self, table_name: str) -> Munch:
        return getattr_rec(self.params.tables, table_name)

    def _set_table_params(self, table_name: str, **fields):
        if table_name in self.all_tables:
            getattr_rec(
                self.params.tables,
                table_name,
            ).update(munchify(fields))
        else:
            setattr_rec(
                self.params.tables,
                table_name,
                munchify(fields),
            )

    @property
    def list_namespaces(self) -> list[str]:
        if self._namespace is not None:
            return [self._namespace]
        else:
            return list(self.params.tables.keys())

    @property
    def namespace(self) -> str | None:
        return self._namespace

    @property
    def all_tables(self) -> list[str]:
        return [
            t.split(".path")[0]
            for t in get_nested_keys(self.params.tables)
            if "path" in t
        ]

    def show(self, table_name: str) -> None:
        print(toYAML(self.get(table_name)))

    def get(self, table_name: str) -> Munch:
        return getattr_rec(self.params.tables, table_name)

    @property
    def all_filesystems(self) -> list[str]:
        return sorted(self.params.filesystem.keys())

    def show_filesystem(self, table_name: str) -> None:
        print(toYAML(self.params.filesystem[table_name]))

    def files(self, table_name: str) -> list[str]:
        params = self._get_table_params(table_name=table_name)
        return sorted(
            self.fs[params.filesystem].glob(params.path + f"/**/*.{params.format}")
        )

    def load_parquet(
        self, table_name: str, as_dataset=True, with_metadata: bool = True, **kwargs
    ) -> ParquetDataset | PyarrowDataset | pl.DataFrame | None:
        params = self._get_table_params(table_name=table_name)

        if "parquet" not in params.format.lower():
            return
        if not as_dataset:
            if params.path.endswith(".parquet"):
                df = self.fs[params.filesystem].read_parquet(params.path, **kwargs)
                self.ddb_con.register(table_name, df)
                return df

            df = self.fs[params.filesystem].read_parquet_dataset(params.path, **kwargs)
            self.ddb_con.register(table_name, df)
            return df

        if with_metadata:
            return ParquetDataset(
                params.path,
                filesystem=self.fs[params.filesystem],
                name=table_name,
                ddb_con=self.ddb_con,
                **kwargs,
            )

        return PyarrowDataset(
            params.path,
            filesystem=self.fs[params.filesystem],
            name=table_name,
            ddb_con=self.ddb_con,
            **kwargs,
        )

    def load_csv(
        self, table_name: str, as_dataset: bool = True, **kwargs
    ) -> CsvDataset | pl.DataFrame | None:
        params = self._get_table_params(table_name=table_name)

        if "csv" not in params.format.lower():
            return
        if not as_dataset:
            if params.path.endswith(".csv"):
                df = self.fs[params.filesystem].read_parquet(params.path, **kwargs)
                self.ddb_con.register(table_name, df)
                return df

            df = self.fs[params.filesystem].read_parquet_dataset(params.path, **kwargs)
            self.ddb_con.register(table_name, df)
            return df

        return CsvDataset(
            params.path,
            filesystem=self.fs[params.filesystem],
            name=table_name,
            ddb_con=self.ddb_con,
            **kwargs,
        )

    def load_json(
        self, table_name: str, as_dataset: bool = True, **kwargs
    ) -> JsonDataset | pl.DataFrame | None:
        params = self._get_table_params(table_name=table_name)

        if "json" not in params.format.lower():
            return
        if not as_dataset:
            if params.path.endswith(".json"):
                df = self.fs[params.filesystem].read_json(params.path, **kwargs)
                self.ddb_con.register(table_name, df)
                return df

            df = self.fs[params.filesystem].read_json_dataset(params.path, **kwargs)
            self.ddb_con.register(table_name, df)
            return df
        return JsonDataset(
            params.path,
            filesystem=self.fs[params.filesystem],
            name=table_name,
            ddb_con=self.ddb_con,
            **kwargs,
        )

    def load(
        self,
        table_name: str,
        as_dataset: bool = True,
        with_metadata: bool = True,
        reload: bool = False,
        **kwargs,
    ):
        params = self._get_table_params(table_name=table_name)

        if params.format.lower() == "parquet":
            if table_name not in self.table and not reload:
                self.table[table_name] = self.load_parquet(
                    table_name,
                    as_dataset=as_dataset,
                    with_metadata=with_metadata,
                    **kwargs,
                )
            return self.table[table_name]

        elif params.format.lower() == "csv":
            if table_name not in self.table and not reload:
                self.table[table_name] = self.load_csv(
                    table_name, as_dataset=as_dataset, **kwargs
                )
            return self.table[table_name]

        elif params.format.lower() == "json":
            if table_name not in self.table and not reload:
                self.table[table_name] = self.load_json(table_name, **kwargs)
            return self.table[table_name]

        # return None

    # def _ddb_table_mapping(self, table_name: str):
    #     params = getattr_rec(self._catalog, self._get_table_from_table_name(table_name=table_name))

    #     return {table_name: [params.path, params.format, params.hive_partitioning]}

    @property
    def registered_tables(self):
        return self.ddb_con.sql("SHOW TABLES")

    def sql(
        self, sql: str, as_dataset: bool = True, with_metadata: bool = True, **kwargs
    ):
        table_names = get_table_names(sql)
        registerd_tables = list(map(lambda x: x[0], self.registered_tables.fetchall()))

        for name in table_names:
            if name not in registerd_tables:
                self.load(
                    name, as_dataset=as_dataset, with_metadata=with_metadata, **kwargs
                )

        return self.ddb_con.sql(sql)

    def create_namespace(self, name: str):
        if name in self.params.tables:
            self.params.tables[name] = Munch()

    def create_table(
        self,
        data: pl.DataFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | duckdb.DuckDBPyRelation
        | ParquetDataset
        | PydalaTable
        | CsvDataset
        | JsonDataset
        | PyarrowDataset
        | None = None,
        table_name: str | None = None,
        namespace: str | None = None,
        overwrite: bool = False,
        write_catalog: bool = True,
        write_args: dict | None = None,
        **fields,
    ):
        if isinstance(data, ParquetDataset | PyarrowDataset | CsvDataset | JsonDataset):
            if table_name is None:
                table_name = data.name

            fields["timestamp_column"] = fields.get(
                "timestamp_column", data._timestamp_column
            )
            fields["path"] = fields.get("path", data._path)
            fields["format"] = fields.get("format", data._format)
            fields["partitioning"] = fields.get("partitioning", data._partitioning)
            fields["partitioning_columns"] = fields.get(
                "partitioning_columns", data._partitioning_columns
            )

        if table_name is None:
            raise Exception("table_name is required")

        if namespace is not None and namespace not in table_name:
            table_name = ".".join(
                namespace.split(".") + table_name.split(".")
            )  # f"{namespace}.{table_name}"

        if table_name in self.all_tables:
            if not overwrite:
                raise Exception(
                    f"Table {table_name} already exists. Use overwrite=True to overwrite or use `write_table` to write "
                    "new data (mode='append' or 'delta') to the table."
                )

        if write_args is not None:
            fields["write_args"] = write_args

        self._set_table_params(table_name=table_name, **fields)

        if write_catalog:
            self._write_catalog()

        if data is not None:
            if isinstance(
                data, ParquetDataset | PyarrowDataset | CsvDataset | JsonDataset
            ):
                if fields.get("path") == data._path:
                    return
            self.write_table(data, table_name, **write_args)

    def delete_table(
        self,
        table_name: str | None = None,
        namespace: str | None = None,
        write_catalog: bool = True,
        vacuum: bool = False,
    ):
        if namespace is not None:
            if table_name is None:
                table_name = namespace
            else:
                if namespace not in table_name:
                    table_name = f"{namespace}.{table_name}"

        if table_name in self.all_tables:
            if vacuum:
                self.load(table_name).vacuum()

            delattr_rec(self.params.tables, table_name)
            if self._namespace is not None:
                delattr_rec(self.params[self._namespace].tables, table_name)
            else:
                delattr_rec(self.params.tables, table_name)

            if write_catalog:
                self._write_catalog(delete_table=table_name)

    def update(
        self,
        table_name: str,
        write_catalog: bool = True,
        write_args: dict | None = None,
        **fields,
    ):
        self.load_catalog()
        if write_args is not None:
            fields["write_args"] = write_args
        self._set_table_params(table_name=table_name, **fields)

        if write_catalog:
            self._write_catalog()

    def write_table(
        self,
        data: pl.DataFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | duckdb.DuckDBPyRelation
        | ParquetDataset
        | PydalaTable
        | CsvDataset
        | JsonDataset
        | PyarrowDataset,
        table_name: str,
        as_dataset: bool = True,
        with_metadata: bool = True,
        update_catalog: bool = False,
        **kwargs,
    ):
        params = self._get_table_params(table_name=table_name)

        if kwargs:
            if "write_args" in params:
                params.write_args.update(kwargs)
            else:
                params.write_args = kwargs
            if update_catalog:
                self.update(table_name=table_name, write_args=params.write_args)

        if as_dataset:
            if table_name not in self.table:
                self.load(table_name=table_name, with_metadata=with_metadata)

            self.table[table_name].write_to_dataset(data, **params.write_args)
        else:
            # raise("Not implemented yet. You can use ")
            if params.format.lower() == "parquet":
                self.fs[params.filesystem].write_parquet(data, params.path, **kwargs)
            elif params.format.lower() == "csv":
                self.fs[params.filesystem].write_csv(data, params.path, **kwargs)
            elif params.format.lower() == "json":
                self.fs[params.filesystem].write_json(data, params.path, **kwargs)

    def schema(self, table_name: str):
        self.load(table_name).schema
